#include <libxml++/libxml++.h>
#include <cassert>
#include <iostream>
#include <set>
#include <fstream>
#include <list>

//! convert string to upper case
std::string
toupper(std::string s) {
     for (unsigned i=0; i<s.length(); ++i) s[i]=toupper(s[i]);
     return s;
}

//! replace all characters f by r in string s
std::string
replace(std::string s, char f, char r) {
     for (unsigned i=0; i<s.length(); ++i) if (s[i]==f) s[i]=r;
     return s;
}

//! strip prefix from string s
/*!
  \return string s without prefix or an empty string on error
 */
std::string
strip(std::string s, std::string prefix)
{
     std::string error;
     if ( (prefix.length()>s.length() ) || ( std::string(s,0,prefix.length())!=prefix ) ) return error;
     return std::string(s, prefix.length(), s.length()-prefix.length());
}

//! extract group from attributes
std::string
extract_group(const std::string &attrs)
{
     // todo: improve this
     std::string error;
     std::string to_match("gccxml(libt2n-");
     std::string::size_type p(attrs.find(to_match));
     if (p==std::string::npos) return error;
     std::string group(attrs, p+to_match.length(), attrs.length());
     p=group.find_first_of(')');
     assert(p!=std::string::npos);
     return std::string(group,0,p);
}

//! get child element by id
/*!
  \return pointer to element having id or null on error
  \todo find libxmlpp pendant
*/
const xmlpp::Element* get_element_by_id(const xmlpp::Element* element, const std::string &id)
{
     const xmlpp::Attribute* cid = element->get_attribute("id");
     if ( cid && ( cid->get_value() == id)) return element;

     //Recurse through child nodes:
     xmlpp::Node::NodeList list = element->get_children();
     for(xmlpp::Node::NodeList::iterator iter = list.begin(); iter != list.end(); ++iter)
     {
	  const xmlpp::Element* element = dynamic_cast<const xmlpp::Element*>(*iter);
	  if (element) {
	       const xmlpp::Element* match = get_element_by_id(element, id);
	       if (match) return match;
	  }
     }
     return NULL;
}

//! get namespace by id
/*!
  \return namespace name or empty string on error
*/
std::string get_namespace(const xmlpp::Element* root, const std::string &id)
{
     std::string error;
     const xmlpp::Element* element(get_element_by_id(root, id));
     if ((!element)||(!element->get_attribute("name"))) return error;
     return element->get_attribute("name")->get_value();
}

//! get default group
std::string get_default_group(const xmlpp::Element* root) {
     std::string error;

     if (   (root->get_attribute("name"))
	 && (root->get_attribute("attributes"))
	 && (root->get_name() == "Typedef")
	 && (root->get_attribute("name")->get_value()=="libt2n_default_group"))
	  return extract_group(root->get_attribute("attributes")->get_value());

     //Recurse through child nodes:
     xmlpp::Node::NodeList list = root->get_children();
     for(xmlpp::Node::NodeList::iterator iter = list.begin(); iter != list.end(); ++iter)
     {
	  if (const xmlpp::Element* elem=dynamic_cast<const xmlpp::Element*>(*iter)) {
	       std::string ret(get_default_group(elem));
	       if (!ret.empty()) return ret;
	  }
     }
     return error;
}

struct type_info
{
     std::string name;
     std::string noref_name;
     bool operator==(const type_info& o) {return (name==o.name) && (noref_name == o.noref_name);}
     std::string noref() const {return noref_name.empty() ? name : noref_name;}
};

std::ostream &operator<<(std::ostream &o, const type_info &t) {
     o << t.name;
     return o;
}

//! get type by id
/*!
  \return type name or empty string on error
*/
type_info get_type(const xmlpp::Element* root, const std::string &id)
{
     type_info error;
     const xmlpp::Element* element(get_element_by_id(root, id));
     if (!element) return error;

     // TODO: not yet complete
     // if we recurse - when do we stop?
     // if it is a typedef? yes? (hmm if the typedef is in the file parsed this will not work)

     // TODO: const and reference types handling is a ugly hack

     std::string tag(element->get_name());
     if (tag=="ReferenceType") {
	  assert(element->get_attribute("type"));
	  type_info ret(get_type(root, element->get_attribute("type")->get_value()));
	  if (ret==error) return error;
	  // at the moment we only support const &
	  // todo: nice error message!
	  if ((ret.noref_name=strip(ret.name,"const ")).empty()) return error;
	  ret.name=ret.name+"&";
	  return ret;
     }else if (tag=="CvQualifiedType") {
	  assert(element->get_attribute("type"));
	  type_info ret(get_type(root, element->get_attribute("type")->get_value()));
	  if (ret==error) return error;
	  ret.name=std::string("const ")+ret.name;
	  return ret;
     }

     assert(element->get_attribute("name"));
     type_info ret;
     if (element->get_attribute("context")) {
	  ret.name=get_namespace(root, element->get_attribute("context")->get_value());
	  if (ret.name!="::")
	       ret.name+="::";
	  else
	       // do not explicitely add ::
	       ret.name="";
     }
     ret.name+=element->get_attribute("name")->get_value();
     return ret;
}

struct t2n_procedure
{
     typedef std::list<std::pair<std::string, type_info> > Args;

     std::string group;
     type_info ret_type;
     std::string name;
     std::string mangled;
     Args  args;

     std::string ret_classname() const {
	  return name+mangled+"_res";
     }
     std::string cmd_classname() const {
	  return name+mangled+"_cmd";
     }
};

std::ostream &operator<<(std::ostream &o, const t2n_procedure::Args &args) {
     for (t2n_procedure::Args::const_iterator it=args.begin();it!=args.end();++it) {
	  if (it!=args.begin()) o << ", ";
	  o << it->second << " " << it->first;
     }
     return o;
}

std::ostream &operator<<(std::ostream &o, const t2n_procedure &f) {
     o << f.ret_type << " " << f.name << "(" << f.args << ")";
     return o;
}

class Parser
{
public:
     Parser(const std::string &fname) : m_fname(fname) {}

     std::list<t2n_procedure> get_procedures() {
	  xmlpp::DomParser parser;
	  //    parser.set_validate();
	  parser.set_substitute_entities(); //We just want the text to be resolved/unescaped automatically.
	  parser.parse_file(m_fname);
	  if(parser)
	  {
	       //Walk the tree:
	       const xmlpp::Node* pNode = parser.get_document()->get_root_node(); //deleted by DomParser.
	       const xmlpp::Element* root = dynamic_cast<const xmlpp::Element*>(pNode);
	       assert(root);
	       visit_node(root);
	  }
	  return m_procedures;
     }
protected:
     std::string m_fname;
     std::list<t2n_procedure> m_procedures;

     void parse_function(const xmlpp::Element* root, const xmlpp::Node* node) {
	  const xmlpp::Element* element = dynamic_cast<const xmlpp::Element*>(node);
	  if (!element) return;

	  const xmlpp::Attribute* attributes = element->get_attribute("attributes");
	  const xmlpp::Attribute* name = element->get_attribute("name");
	  const xmlpp::Attribute* mangled = element->get_attribute("mangled");
	  const xmlpp::Attribute* returns = element->get_attribute("returns");
	  if ((!attributes)||(!name)||(!mangled)||(!returns)) return;

	  // check wether the procedure is marked (TODO: improve)
	  // attributes are speparated by spaces?

	  t2n_procedure f;
	  f.group=extract_group(attributes->get_value());
	  if (f.group.empty()) return;
	  // todo: handle default group
	  if (f.group=="default") f.group=get_default_group(root);

	  // we need the return type
	  f.ret_type=get_type(root, returns->get_value());
	  f.name=name->get_value();
	  f.mangled=mangled->get_value();

	  xmlpp::Node::NodeList list = node->get_children("Argument");
	  for(xmlpp::Node::NodeList::iterator iter = list.begin(); iter != list.end(); ++iter)
	  {
	       const xmlpp::Element* arg = dynamic_cast<const xmlpp::Element*>(*iter);
	       if ( arg ) {
		    assert(arg->get_name() == "Argument");
		    assert(arg->get_attribute("name"));
		    assert(arg->get_attribute("type"));
		    f.args.push_back(std::pair<std::string, type_info>(arg->get_attribute("name")->get_value(), get_type(root, arg->get_attribute("type")->get_value())));
	       }
	  }
	  std::cerr << "Found function: " << f << std::endl;
	  m_procedures.push_back(f);
     }

     void visit_node(const xmlpp::Element* root, const xmlpp::Node* node = NULL, unsigned int indentation = 0)
	  {
	       if (!node) node=root;
	  
	       const xmlpp::ContentNode* nodeContent = dynamic_cast<const xmlpp::ContentNode*>(node);
	       const xmlpp::TextNode* nodeText = dynamic_cast<const xmlpp::TextNode*>(node);
	       const xmlpp::CommentNode* nodeComment = dynamic_cast<const xmlpp::CommentNode*>(node);

	       if(nodeText && nodeText->is_white_space()) //Let's ignore the indenting - you don't always want to do this.
		    return;
    
	       std::string nodename = node->get_name();

	       if(!nodeText && !nodeComment && !nodename.empty()) //Let's not say "name: text".
	       {
		    if (node->get_name() == "Function") parse_function(root, node);
	       }
	       if(!nodeContent)
	       {
		    //Recurse through child nodes:
		    xmlpp::Node::NodeList list = node->get_children();
		    for(xmlpp::Node::NodeList::iterator iter = list.begin(); iter != list.end(); ++iter)
		    {
			 visit_node(root, *iter, indentation + 2); //recursive
		    }
	       }
	  }
};

//! find used groups
std::set<std::string>
used_groups(const std::list<t2n_procedure> &funcs) {
     typedef std::set<std::string> Groups;
     Groups groups;
     for (std::list<t2n_procedure>::const_iterator it=funcs.begin();it!=funcs.end();++it)
	  // since we use std::set each group is inserted only once
	  groups.insert(it->group);
     return groups;
}

void output_common_hpp(std::ostream &o, const std::list<t2n_procedure> &procs) {
     std::set<std::string> groups(used_groups(procs));

     for (std::set<std::string>::iterator it=groups.begin();it!=groups.end();++it) {
	  o << "class cmd_group_" << *it << " : public libt2n::command\n"
	    << "{\n"
	    << "private:\n"
	    << "	friend class boost::serialization::access;\n"
	    << "	template<class Archive>\n"
	    << "	void serialize(Archive & ar, const unsigned int /* version */)\n"
	    << "	{ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(libt2n::command);}\n"
	    << "};\n";
     }
     
     for (std::list<t2n_procedure>::const_iterator it=procs.begin();it!=procs.end();++it) {
	  o << "class " << it->ret_classname() << " : public libt2n::result\n"
	    << "{\n"
	    << "private:\n"
	    << "	" << it->ret_type << " res;\n"
	    << "	friend class boost::serialization::access;\n"
	    << "	template<class Archive>\n"
	    << "	void serialize(Archive & ar, const unsigned int /* version */)\n"
	    << "	{\n"
	    << "		ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(libt2n::result);\n"
	    << "		ar & BOOST_SERIALIZATION_NVP(res);\n"
	    << "	}\n"
	    << "public:\n"
	    << "	" << it->ret_classname() << "() {}\n"
	    << "	" << it->ret_classname() << "(const " << it->ret_type << " &_res) : res(_res) {}\n"
	    << "	" << it->ret_type << " get_data() { return res; }\n"
	    << "};\n";
     }
     for (std::list<t2n_procedure>::const_iterator it=procs.begin();it!=procs.end();++it) {
	  o << "class " << it->cmd_classname() << " : public " << "cmd_group_" << it->group << "\n"
	    << "{\n"
	    << "private:\n";
	  for (t2n_procedure::Args::const_iterator ait=it->args.begin();ait!=it->args.end();++ait) {
	       o << "	" << ait->second.noref() << " " << ait->first << ";\n";
	  }
	  o << "	friend class boost::serialization::access;\n"
	    << "	template<class Archive>\n"
	    << "	void serialize(Archive & ar, const unsigned int /* version */)\n"
	    << "	{\n"
	    << "		ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(cmd_group_" << it->group << ");\n";
	  for (t2n_procedure::Args::const_iterator ait=it->args.begin();ait!=it->args.end();++ait) {
	       o << "		ar & BOOST_SERIALIZATION_NVP(" << ait->first << ");\n";
	  }

	  // default constructor
	  o << "	}\n"
	    << "\n"
	    << "public:\n"
	    << "	" << it->cmd_classname() << "() {}\n";

	  // constructor taking all arguments
	  o << "	" << it->cmd_classname() << "(";
	  for (t2n_procedure::Args::const_iterator ait=it->args.begin();ait!=it->args.end();++ait) {
	       if (ait!=it->args.begin()) o << ", ";
	       o << ait->second << " _" << ait->first;
	  }
	  o << ") : ";
	  for (t2n_procedure::Args::const_iterator ait=it->args.begin();ait!=it->args.end();++ait) {
	       if (ait!=it->args.begin()) o << ", ";
	       o << ait->first << "(_" << ait->first << ")";
	  }
	  o << " {}\n"
	    << "	libt2n::result* operator()();\n"
	    << "};\n";
     }
}

void output_common_cpp(std::ostream &o, const std::list<t2n_procedure> &procs, const std::string &common_hpp) {
     std::set<std::string> groups(used_groups(procs));

     o << "#include \"" << common_hpp << "\"\n"
       << "#include <boost/serialization/export.hpp>\n"
       << "\n"
       << "/* register types with boost serialization */\n";
     for (std::set<std::string>::iterator it=groups.begin();it!=groups.end();++it) {
	  o << "BOOST_CLASS_EXPORT(cmd_group_" << *it << ")\n";
     }
     for (std::list<t2n_procedure>::const_iterator it=procs.begin();it!=procs.end();++it) {
	  o << "BOOST_CLASS_EXPORT("<<it->ret_classname()<<")\n"
	    << "BOOST_CLASS_EXPORT("<<it->cmd_classname()<<")\n";
     }
}

void output_client_hpp(std::ostream &o, const std::list<t2n_procedure> &procs) {
     std::set<std::string> groups(used_groups(procs));

     o << "#include <command_client.hxx>\n";

     for (std::set<std::string>::iterator it=groups.begin();it!=groups.end();++it) {
	  o << "class cmd_group_" << *it << "_client : public libt2n::command_client\n"
	    << "{\n"
	    << "public:\n"
	    << "cmd_group_" << *it << "_client(libt2n::client_connection &_c,\n"
	    << "	long long _command_timeout_usec=command_timeout_usec_default,\n"
	    << "	long long _hello_timeout_usec=hello_timeout_usec_default)\n"
	    << "	: libt2n::command_client(_c,_command_timeout_usec,_hello_timeout_usec)\n"
            << "	{}\n";
	  for (std::list<t2n_procedure>::const_iterator pit=procs.begin();pit!=procs.end();++pit) {
	       if (pit->group==*it) {
		    o << "	" << *pit << ";\n";
	       }
	  }
	  o << "};\n";
     }
}

void output_client_cpp(std::ostream &o, const std::list<t2n_procedure> &procs, const std::string &common_hpp, const std::string &common_cpp, const std::string &client_hpp) {
     std::set<std::string> groups(used_groups(procs));

     o << "#include \"" << client_hpp << "\"\n"
       << "#include \"" << common_hpp << "\"\n"
       << "// fake\n";
     for (std::list<t2n_procedure>::const_iterator it=procs.begin();it!=procs.end();++it) {
	  o << "libt2n::result* " << it->cmd_classname() << "::operator()() { return NULL; }\n";
     }

     for (std::set<std::string>::iterator it=groups.begin();it!=groups.end();++it) {
	  for (std::list<t2n_procedure>::const_iterator pit=procs.begin();pit!=procs.end();++pit) {
	       if (pit->group==*it) {
		    o << pit->ret_type << " cmd_group_" << *it << "_client::" << pit->name << "(" << pit->args << ")\n"
		      << "{\n"
		      << "	libt2n::result_container rc;\n"
		      << "	send_command(new " << pit->cmd_classname() << "(";
		    for (t2n_procedure::Args::const_iterator ait=pit->args.begin();ait!=pit->args.end();++ait) {
			 if (ait!=pit->args.begin()) o << ", ";
			 o << ait->first;
		    }
		    o << "), rc);\n"
		      << "	" << pit->ret_classname() << "* res=dynamic_cast<" << pit->ret_classname() << "*>(rc.get_result());\n"
		      << "	if (!res) throw libt2n::t2n_communication_error(\"result object of wrong type\");\n"
		      << "	return res->get_data();\n"
		      << "}\n";
	       }
	  }
     }

     // include in this compilation unit to ensure the compilation unit is used
     // see also:
     // http://www.google.de/search?q=g%2B%2B+static+initializer+in+static+library
     o << "#include \"" << common_cpp << "\"\n";
}

void output_server_cpp(std::ostream &o, const std::list<t2n_procedure> &procs, const std::string &common_hpp, const std::string &common_cpp) {
     o << "#include \"" << common_hpp << "\"\n";

     for (std::list<t2n_procedure>::const_iterator it=procs.begin();it!=procs.end();++it) {
	  o << *it << ";\n";
	  o << "libt2n::result* " << it->cmd_classname() << "::operator()() { return new " << it->ret_classname() << "(" << it->name << "(";
	  for (t2n_procedure::Args::const_iterator ait=it->args.begin();ait!=it->args.end();++ait) {
	       if (ait!=it->args.begin()) o << ", ";
	       o << ait->first;
	  }
	  o << ")); }\n";
     }
     o << "#include \"" << common_cpp << "\"\n";
}

struct header_file : public std::ofstream
{
     header_file(const char* fname) : std::ofstream(fname) {
	  std::cerr << "create header: '" << fname << "'" << std::endl;
	  std::string macro(replace(toupper(fname),'.','_'));
	  *this << "// automatically generated code - do not edit\n" << std::endl;
	  *this << "#ifndef " << macro << "\n"
		<< "#define " << macro << "\n";
     }
     ~header_file() {
	  *this << "#endif" << std::endl;
     }
};

struct cpp_file : public std::ofstream
{
     cpp_file(const char* fname) : std::ofstream(fname) {
	  std::cerr << "create cpp: '" << fname << "'" << std::endl;
	  *this << "// automatically generated code - do not edit\n" << std::endl;
     }
};

std::list<std::string>
get_includes(const std::string &fname)
{
     // grep "#include" fname
     std::ifstream in(fname.c_str());
     std::string line;
     std::list<std::string> ret;
     while (std::getline(in,line)) {
	  if (line.find("#include")!=std::string::npos)
	       ret.push_back(line);
     }
     return ret;
}

void
paste_includes(std::ostream &o, std::list<std::string> &i)
{
     o << std::endl
       << "// copied includes begin" << std::endl;
     for (std::list<std::string>::const_iterator it=i.begin(); it!=i.end(); ++it)
	  o << *it << std::endl;
     o << "// copied includes end" << std::endl
       << std::endl;
}

struct RemoveGenerated
{
     RemoveGenerated(const std::string &_prefix) : prefix(_prefix) {}
     bool operator()(const std::string &s) const {
	  return (s.find(prefix+"common.hxx")!=std::string::npos);
     }
     const std::string &prefix;
};

int
main(int argc, char* argv[])
{
     if (argc != 4) {
	  std::cerr << "Usage: " << argv[0] << " header-file gccxml-file outputprefix" << std::endl;
	  return 1;
     }
     std::string headerfile(argv[1]);
     std::string xmlfile(argv[2]);
     std::string prefix(argv[3]);
     std::list<std::string> includes(get_includes(headerfile));
     remove_if(includes.begin(), includes.end(), RemoveGenerated(prefix));
     includes.erase(remove_if(includes.begin(), includes.end(), RemoveGenerated(prefix)), includes.end());

     paste_includes(std::cerr, includes);
  
     Parser parser(xmlfile);
     std::list<t2n_procedure> procedures(parser.get_procedures());

     std::cerr << "Procedures:" << std::endl;
     for (std::list<t2n_procedure>::const_iterator it=procedures.begin();it!=procedures.end();++it)
	  std::cerr << *it << ";" << std::endl;

     std::set<std::string> groups(used_groups(procedures));
     std::cerr << "Used groups:" << std::endl;
     for (std::set<std::string>::const_iterator it=groups.begin();it!=groups.end();++it)
	  std::cerr << *it << std::endl;

     std::string common_hpp_fname(prefix+"common.hxx");
     std::string common_cpp_fname(prefix+"common.cpp");
     std::string client_hpp_fname(prefix+"client.hxx");
     std::string client_cpp_fname(prefix+"client.cpp");
     std::string server_cpp_fname(prefix+"server.cpp");

     header_file common_hpp(common_hpp_fname.c_str());
     common_hpp << "#include \"codegen-stubhead.hxx\"\n";
     paste_includes(common_hpp, includes);

     output_common_hpp(common_hpp, procedures);

     cpp_file common_cpp(common_cpp_fname.c_str());
     output_common_cpp(common_cpp, procedures, common_hpp_fname);

     header_file client_hpp(client_hpp_fname.c_str());
     client_hpp << "#include \"codegen-stubhead.hxx\"\n";
     // we can't paste the includes before codegen-stubhead.hxx was included
     paste_includes(client_hpp, includes);
     output_client_hpp(client_hpp, procedures);

     cpp_file client_cpp(client_cpp_fname.c_str());
     output_client_cpp(client_cpp, procedures, common_hpp_fname, common_cpp_fname, client_hpp_fname);

     cpp_file server_cpp(server_cpp_fname.c_str());
     output_server_cpp(server_cpp, procedures, common_hpp_fname, common_cpp_fname);
     return 0;
}
